from astropy.coordinates import SkyCoord
import astropy.units as u
import xarray as xr
import logging


class Field:
    def __init__(self, _id: int = None, dataset: xr.Dataset = None):
        """
        Class that represents each field in the observation.

        Parameters
        ----------
        _id : int
             id of the field Object
        dataset : xarray.Dataset
                 xarray dataset of the field object
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self._id = _id
        self.dataset = dataset

        self.ref_dir = None
        self.phs_dir = None

        self.name = ""
        if dataset is not None and _id is not None:
            ref_dir = self.dataset.REFERENCE_DIR.data.squeeze().compute()
            self.ref_dir = SkyCoord(ra=ref_dir[0] * u.rad, dec=ref_dir[1] * u.rad)

            phs_dir = self.dataset.PHASE_DIR.data.squeeze().compute()
            self.phs_dir = SkyCoord(ra=phs_dir[0] * u.rad, dec=phs_dir[1] * u.rad)

            self.name = self.dataset.NAME.data.squeeze().compute()



from abc import ABC
import logging
import xarray
import dask.array as da
import numpy as np
import dask
from daskms import xds_from_table, xds_from_ms, xds_to_table
from pyralysis.io.io import Io
from ..base.antenna import Antenna
from ..base.baseline import Baseline
from ..base.dataset import Dataset
from ..base.field import Field
from ..base.ms import MS
from ..base.polarization import Polarization
from ..base.spectral_window import SpectralWindow
from ..base.visibility import Visibility
from typing import Union, List
import shutil
import os


class DaskMS(Io, ABC):
    def __init__(self, **kwargs):
        """
        I/O class to handle Measurements Sets v2 using Dask
        Parameters
        ----------
        kwargs : IO arguments
        """
        super().__init__(**kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.ms_name_dask = self.input_name + "::"

    # This should look for CORRECTED_DATA and if does not exist then look for DATA
    def read(self, read_flagged_data: bool = False, datacolumn: str = None) -> Dataset:
        """
        Function that reads a Measurement Set v2 lazily using Dask.

        Parameters
        ----------
        read_flagged_data : Whether to read flagged_data or not.
        datacolumn : Data column to read. E.g: "corrected", "model", "data"

        Returns
        -------
        Returns a Pyralysis Dataset object
        """

        taql_query = ""
        if not read_flagged_data:
            taql_query = "!FLAG_ROW"

        # Creating antenna object
        antennas = xds_from_table(self.ms_name_dask + "ANTENNA", taql_where=taql_query)[0]
        antenna_obj = Antenna(dataset=antennas)

        # Creating baseline object
        baseline_obj = Baseline(antenna_obj)

        # Reading spectral window datasets (separated by different spectral windows)
        spws = xds_from_table(self.ms_name_dask + "SPECTRAL_WINDOW", taql_where=taql_query, group_cols="__row__")

        # Reading field datasets (separated by different fields)
        fields = xds_from_table(self.ms_name_dask + "FIELD", taql_where=taql_query, group_cols="__row__")

        # Reading data description dataset
        ddids = xds_from_table(self.ms_name_dask + "DATA_DESCRIPTION", taql_where=taql_query)[0]

        # Reading polarization datasets (separated by different polarization rows)
        pols = xds_from_table(self.ms_name_dask + "POLARIZATION", taql_where=taql_query, group_cols="__row__")

        # Reading MS main table indexing columns such as SCAN_NUMBER, TIME, ANTENNA1 and ANTENNA2
        # This returns a list of measurement sets partitioned by FIELDS and SPWS
        ms = xds_from_ms(self.input_name, taql_where=taql_query,
                         index_cols=["SCAN_NUMBER", "TIME", "ANTENNA1", "ANTENNA2"])

        # Check if datacolumn exists in the measurement set
        # Getting keys of the dataset
        ms_keys = ms[0].data_vars.keys()
        if datacolumn is None:
            if "CORRECTED_DATA" in ms_keys and "DATA" in ms_keys:
                datacolumn = "CORRECTED_DATA"
            elif "CORRECTED_DATA" in ms_keys:
                datacolumn = "CORRECTED_DATA"
            elif "DATA" in ms_keys:
                datacolumn = "DATA"
            else:
                self.logger.error("There is no valid column to read")
                raise ValueError
        else:
            if datacolumn not in ms_keys:
                self.logger.error("Column " + datacolumn + " does not exist")
                raise ValueError

        ms_list = []
        ms_id = 0
        for msds in ms:
            fdid = msds.attrs['FIELD_ID']
            ddid = msds.attrs['DATA_DESC_ID']
            spw_id = ddids["SPECTRAL_WINDOW_ID"][ddid].data.compute()
            pol_id = ddids["POLARIZATION_ID"][ddid].data.compute()
            field = fields[fdid]  # This is a field dataset
            spw = spws[spw_id]  # This is a spw dataset
            pol = pols[pol_id]  # This is a polarization dataset
            row_id = msds["ROWID"].data  # Row ID
            uvw = msds["UVW"].data  # UVW
            data = msds[datacolumn].data  # DATA
            weight = msds["WEIGHT"].data  # WEIGHT
            flag = msds["FLAG"].data  # FLAGS
            antenna1_ids = msds["ANTENNA1"].data  # ANTENNA 1 ID
            antenna2_ids = msds["ANTENNA2"].data  # ANTENNA 2 ID
            baseline_id = da.argmax((baseline_obj.dataset.ANTENNA1.data[None, :] == antenna1_ids[:, None]) & (
                    baseline_obj.dataset.ANTENNA2.data[None, :] == antenna2_ids[:, None]), axis=1).astype(np.int32)

            ds = xarray.Dataset(
                data_vars=dict(
                    UVW=(["row", "uvw"], uvw),
                    DATA=(["row", "chan", "corr"], data),
                    WEIGHT=(["row", "corr"], weight),
                    FLAG=(["row", "chan", "corr"], flag),
                ),
                coords=dict(
                    ROWID=(["row"], row_id),
                    ANTENNA1=(["row"], antenna1_ids),
                    ANTENNA2=(["row"], antenna2_ids),
                    BASELINE=(["row"], baseline_id)
                )
            )

            field_obj = Field(_id=fdid, dataset=field)
            spw_obj = SpectralWindow(_id=spw_id, dataset=spw)
            pol_obj = Polarization(_id=pol_id, dataset=pol)
            visibility_obj = Visibility(dataset=ds)
            ms_obj = MS(_id=ms_id, field=field_obj, polarization=pol_obj, spectral_window=spw_obj,
                        visibilities=visibility_obj)
            ms_list.append(ms_obj)
            ms_id += 1
        return Dataset(antenna=antenna_obj, baseline=baseline_obj, spectral_window_dataset=spws, ms_list=ms_list)

    def write_xarray_ds(self, dataset: Union[xarray.Dataset, List[xarray.Dataset]] = None, ms_name: str = None,
                        table_name: str = None,
                        columns: Union[List[str], str] = None) -> None:
        """
        Function that writes a xarray Dataset to a Measurement Set File
        Parameters
        ----------
        dataset : xarray Dataset to write
        ms_name : Name of the output Measurement Set
        table_name : Name of the table you might write. Default: None
        columns : Name of the columns to write to the table
        """
        if dataset is None:
            raise ValueError("Cannot write into Measurement Set without an xarray Dataset object")

        if ms_name is None:
            ms_name = self.input_name

        if table_name is not None:
            ms_name = ms_name + "::" + table_name

        if columns is None:
            columns = "ALL"

        writes = xds_to_table(dataset, table_name=ms_name, columns=columns)  # TODO: This could be returned and then
        # the compute can be done by the user
        dask.compute(writes)

    def write(self, dataset: Dataset = None, ms_name: str = None, columns: Union[List[str], str] = None) -> None:
        """
        Function that writes a Pyralysis Dataset (only ms_list) to a Measurement Set File
        Parameters
        ----------
        dataset : Pyralysis Dataset to write
        ms_name : Name of the output Measurement Set
        columns : Name of the columns to write to the table
        """
        if ms_name is None:
            ms_name = self.input_name
        else:
            if self.input_name != ms_name:
                if os.path.exists(ms_name):
                    shutil.rmtree(ms_name)
                shutil.copytree(self.input_name, ms_name)

        ms_datasets = [ms.visibilities.dataset for ms in dataset.ms_list]

        self.write_xarray_ds(dataset=ms_datasets, ms_name=ms_name, columns=columns)



from .transformer import Transformer
from .gridder import Gridder
from abc import ABC, ABCMeta, abstractmethod
import copy


class WeightingScheme(Transformer, metaclass=ABCMeta):
    def __init__(self, gridder: Gridder = None, **kwargs):
        """
        Class that represents the interferometry weighting scheme
        :param gridding: The gridding object in order to grid weights if uniform or briggs weighting
        :param kwargs: Transformer object arguments
        """
        super().__init__(**kwargs)
        self.gridder = gridder

    @abstractmethod
    def transform(self):
        """
        Transforms the weights of an input dataset
        """
        pass



from abc import ABC
from ..utils.analytical_functions import Gaussian2D


class PSF(Gaussian2D, ABC):
    def __init__(self, **kwargs):
        """
        Class that represents the Point Spread Function. Inherits from Gaussian 2D.
        :param kwargs: Gaussian2D arguments
        """
        super().__init__(**kwargs)



import astropy.units as u
import astropy.constants as co
from typing import Union
from astropy.units import Quantity
import dask.array as da
import xarray as xr
from .units_functions import dask_unit_conversion, xarray_unit_conversion
from astropy.units.equivalencies import Equivalency

lambdas = u.def_unit('lambdas', format={'format': r'\lambda'}, prefixes=True, namespace=u.__dict__)


def lambdas_equivalencies(restfreq: Union[Quantity, da.core.Array, xr.DataArray] = None) -> Equivalency:
    """
    Creates the equivalencies given a rest frequency
    :param restfreq: Rest Frequency (can be a Quantity or an array)
    :return: Returns the equivalencies
    """
    if restfreq is not None:
        if isinstance(restfreq, Quantity):
            restfreq_hz = restfreq.to(u.Hz, equivalencies=u.spectral())
            restfreq_to_m = restfreq_hz.to(u.m, u.spectral())
        elif isinstance(restfreq, da.core.Array):
            restfreq_hz = dask_unit_conversion(restfreq, u.Hz, u.spectral())
            restfreq_to_m = dask_unit_conversion(restfreq, u.m, u.spectral())
        elif isinstance(restfreq, xr.DataArray):
            restfreq_hz = xarray_unit_conversion(restfreq, u.Hz, u.spectral())
            restfreq_to_m = xarray_unit_conversion(restfreq, u.m, u.spectral())
        else:
            raise ValueError("This equivalency only accepts Quantity, dask.array or xr.DataArray instances")

    # Beware that if restfreq is None then you cannot convert between lambdas-seconds or lambdas-meters

    eq = [
        (lambdas, u.s, lambda x: x / restfreq_hz, lambda x: x * restfreq_hz),
        (u.lambdas, u.m, lambda x: x * restfreq_to_m, lambda x: x / restfreq_to_m),
        (u.m, u.s, lambda x: x / co.c.to(u.m / u.s).value, lambda x: x * co.c.to(u.m / u.s).value),
        (u.lambdas, u.rad, lambda x: 1.0 / x, lambda x: 1.0 / x),
    ]

    return eq



from astropy.units import beam_angular_area
from astropy.units import Quantity
from astropy.units import Unit
from .units_functions import check_units
from typing import Union
import numpy as np


def beam_equivalencies(beam_area: Quantity = None, bmaj: Union[Quantity, float] = None, bmin: Union[Quantity, float] = None, unit: Unit = None):
    if beam_area is None and bmaj is not None and bmin is not None:
        b_area = np.pi * bmaj * bmin / (4. * np.log(2))
    elif beam_area is None and bmaj is None and bmin is None:
        raise ValueError("Input parameters cannot be None")

    if isinstance(beam_area, float) and unit is not None:
        b_area *= unit
    elif isinstance(beam_area, Quantity) and unit is not None:
        if beam_area.unit is None or beam_area.unit.is_unity():
            beam_area *= unit
        else:
            raise ValueError("Area cannot be unitless")
    else:
        raise ValueError("Area cannot be unitless")
    return beam_angular_area(beam_area)



import logging
import astropy.units as u
import xarray as xr


class SpectralWindow:
    def __init__(self, _id: int = None, dataset: xr.Dataset = None):
        """
        Class that represents the Spectral Window table for the partitioned Measurement Set

        Parameters
        ----------
        _id : int
              ID of the Spectral Window
        dataset : xarray.Dataset
                  xarray Dataset representing the Spectral Window table
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self._id = _id
        self.dataset = dataset

        self.nchans = 0
        self.chans = []
        self.chan_width = []
        self.total_bandwidth = 0.0 * u.Hz
        self.ref_frequency = 0.0 * u.Hz
        if dataset is not None and _id is not None:
            self.nchans = int(dataset.NUM_CHAN.data.squeeze().compute())
            self.chans = dataset.CHAN_FREQ.data.squeeze(axis=0) * u.Hz
            self.chan_width = dataset.CHAN_WIDTH.data.squeeze(axis=0).compute() * u.Hz
            self.total_bandwidth = dataset.TOTAL_BANDWIDTH.data.squeeze().compute() * u.Hz
            self.ref_frequency = dataset.REF_FREQUENCY.data.squeeze().compute() * u.Hz



from abc import ABC
from ..parameter import Parameter
import dask.array as da
import xarray as xr
import numpy as np
from typing import List, Union


class CKernel(ABC):
    def __init__(self, amp: float = None, center=List[float], size=Union[List[int], int],
                 w: float = None, dask_array: bool = False):
        """
        Class that represents the convolution kernel
        :param amp: Amplitude of the kernel
        :param center: Center of the kernel
        :param size: Size of image containing the kernel
        :param w: Size of the kernel
        :param dask_array: Whether to store the kernel in a dask array
        """
        self.amp = amp
        self.x0 = center[0]
        self.y0 = center[1]
        self.support_x = None
        self.support_y = None
        if self.data is not None and self.data.ndims > 2:
            self.m = self.data.shape[0]
            self.n = self.data.shape[1]
            self.support_x = self.m // 2
            self.support_y = self.n // 2
        elif self.data is not None and size is not None:
            if isinstance(size, list):
                self.m = size[0]
                self.n = size[1]
            else:
                self.m = size
                self.n = size
            self.support_x = self.m // 2
            self.support_y = self.m // 2
            if dask_array:
                self.data = xr.DataArray(da.zeros((self.m, self.n)))
                self.data.data.compute_chunk_sizes()
            else:
                self.data = xr.DataArray(np.zeros((self.m, self.n)))

        self.w = w
        self.gcf = CKernel()



from .transformer import Transformer
from abc import ABC, ABCMeta, abstractmethod
import xarray as xr
import astropy.units as un
import dask.array as da


class HermitianSymmetry(Transformer, metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
               Transformer object arguments
        """
        super().__init__(**kwargs)

    def transform(self):
        """
        Applies hermitian symmetry to the dataset
        """
        for ms in self.input_data.ms_list:
            bool_array = xr.apply_ufunc(lambda x: x.value < 0.0, ms.visibilities.uvw[:, 0], dask="parallelized",
                                        output_dtypes=[bool])
            # This is faster than using the indexes as dask arrays
            idx = da.argwhere(bool_array).squeeze().compute()
            if len(idx) > 0:
                ms.visibilities.uvw = xr.where(bool_array, ms.visibilities.uvw * -1.0, ms.visibilities.uvw)
                ms.visibilities.data[idx] = da.conj(ms.visibilities.data[idx])
                ms.visibilities.model[idx] = da.conj(ms.visibilities.data[idx])



from abc import ABC

from ..wscheme import WeightingScheme


class Natural(WeightingScheme, ABC):
    def __init__(self, **kwargs):
        """
        Class that represents the Natural Weighting Scheme.
        :param kwargs: WeightingScheme arguments
        """
        super().__init__(**kwargs)
        self.backup = False

    def transform(self):
        """
        For natural weights the weights are not transform, so this function does nothing.
        """
        pass



import numpy as np
from numba import vectorize
from abc import ABCMeta, abstractmethod
import astropy.units as u
from typing import List, Union
from astropy.units import Quantity
import dask.array as da
import xarray as xr
from ..units.units_functions import check_units


@vectorize
def evaluate_gaussian1D(x, amp, mu, sigma):
    return amp * np.exp(-0.5 * ((x - mu) / sigma) ** 2)


class Gaussian1D(metaclass=ABCMeta):
    def __init__(self, amplitude: float = None, mu: float = None, sigma: float = None, fwhm: float = None):
        """
        Class that represents a 1D Gaussian
        :param amplitude: Amplitude of the Gaussian
        :param mu: Center of the Gaussian
        :param sigma: Standard deviation of the Gaussian
        :param fwhm: FWHM of the Gaussian
        """
        if amplitude is not None:
            self.amplitude = amplitude
        else:
            self.amplitude = 1.0

        if mu is not None:
            self.mu = mu
        else:
            self.mu = 0.0

        if fwhm is None and sigma is not None:
            self.sigma = sigma
        elif fwhm is not None:
            self.fwhm = fwhm

        if sigma is None and fwhm is None:
            self.sigma = 1.0

    @property
    def sigma(self):
        return self.__sigma

    @sigma.setter
    def sigma(self, val):
        self.__sigma = val
        val_fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0))
        self.__fwhm = val * val_fwhm

    @property
    def fwhm(self):
        return self.__fwhm

    @fwhm.setter
    def fwhm(self, val):
        self.__fwhm = val
        val_fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0))
        self.__sigma = val / val_fwhm

    def evaluate(self, x=None):
        if x is not None:
            return evaluate_gaussian1D(x, self.amplitude, self.mu, self.sigma)


class Gaussian2D(metaclass=ABCMeta):
    def __init__(self, amplitude: float = None, mu: Union[Quantity, float, List[float]] = None,
                 sigma: Union[Quantity, float, List[float]] = None, theta: Union[Quantity, float] = None,
                 fwhm: Union[Quantity, float, List[float]] = None):
        """
        Class that represents an analytical model of a 2D Elliptical Gaussian
        :param amplitude: Amplitude of the Gaussian
        :param mu: Center of the Gaussian
        :param sigma: Standard deviation of the Gaussian
        :param theta: Inclination angle of the Gaussian
        :param fwhm: FWHM of the Gaussian
        """
        if amplitude is not None:
            self.amplitude = amplitude
        else:
            self.amplitude = 1.0

        self.mu = mu

        if fwhm is None and sigma is not None:
            self.sigma = sigma
        elif fwhm is not None:
            self.fwhm = fwhm

        if sigma is None and fwhm is None:
            self.sigma = [1.0, 1.0]

        if theta is None:
            self.theta = 0.0 * u.rad
        else:
            if isinstance(theta, float):
                self.theta = theta * u.rad
            else:
                # Make sure theta is in radians -> Convert theta to radians
                self.theta = theta.to(u.rad)

    @property
    def mu(self):
        return self.__mu

    @mu.setter
    def mu(self, val):
        if val is not None:
            if isinstance(val, Quantity):
                if isinstance(val.value, float):
                    self.__mu = [val, val]
                else:
                    self.__mu = val
            elif isinstance(val, float):
                self.__mu = [val, val]
            else:
                self.__mu = val
        else:
            self.__mu = [0.0, 0.0]

    @property
    def sigma(self):
        return self.__sigma

    @sigma.setter
    def sigma(self, val):
        if val is not None:
            if isinstance(val, Quantity):
                if isinstance(val.value, float):
                    self.__sigma = [val, val]
                else:
                    self.__sigma = val
            elif isinstance(val, float):
                self.__sigma = [val, val]
            else:
                self.__sigma = val
        else:
            self.__sigma = [1.0, 1.0]
        self.__area = self.calculate_area()
        self.__fwhm = [2.0 * np.sqrt(2.0 * np.log(2.0)) * self.__sigma[0],
                       2.0 * np.sqrt(2.0 * np.log(2.0)) * self.__sigma[1]]

    @property
    def fwhm(self):
        return self.__fwhm

    @fwhm.setter
    def fwhm(self, val):
        self.__fwhm = val
        val_fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0))
        if isinstance(val, Quantity):
            if val.size == 1:
                self.__sigma = Quantity([val / val_fwhm, val / val_fwhm])
            else:
                self.__sigma = Quantity([val[0] / val_fwhm, val[1] / val_fwhm])
        else:
            if isinstance(val, list):
                if len(val) == 2:
                    self.__sigma = [val[0] / val_fwhm, val[1] / val_fwhm]
                else:
                    raise ValueError("The length of the list must not be greater than zero")
            else:
                self.__sigma = [val / val_fwhm, val / val_fwhm]
        self.__area = self.calculate_area()

    @property
    def area(self):
        return self.__area

    def calculate_area(self):
        area = np.pi * self.sigma[0] * self.sigma[1] / (4. * np.log(2))
        if check_units(area, u.steradian):
            return area.to(u.steradian)

    def evaluate(self, x: Union[da.core.Array, xr.DataArray, np.ndarray, float] = None,
                 y: Union[da.core.Array, xr.DataArray, np.ndarray, float] = None):
        """
        Function that evaluates an x,y coordinate in the 2D Gaussian
        :param x: x coordinate
        :param y: y coordinate
        :return: Depending of the input it will return the evaluation in all the (x,y) coordinates
        """
        if x is not None and y is not None:
            a = np.cos(self.theta) ** 2 / self.sigma[0] ** 2 + np.sin(self.theta) ** 2 / self.sigma[1] ** 2
            b = np.sin(2.0 * self.theta) / self.sigma[0] ** 2 + np.sin(2.0 * self.theta) / self.sigma[1] ** 2
            c = np.sin(self.theta) ** 2 / self.sigma[0] ** 2 + np.cos(self.theta) ** 2 / self.sigma[1] ** 2
            exp_factor = -0.5 * (
                        a * (x - self.mu[0]) ** 2 + b * (x - self.mu[0]) * (y - self.mu[1]) + c * (y - self.mu[1]) ** 2)
            f_gauss = self.amplitude * np.exp(exp_factor)
            return f_gauss



from abc import ABCMeta, abstractmethod
import copy
import sys
from ..base import Dataset


class Transformer(metaclass=ABCMeta):
    def __init__(self, input_data: Dataset = None, backup: bool = True):
        """
        Class that represents a transformer. The instance of this class must perform
        a transformation to the input dataset
        :param input_data: The input dataset
        :param backup: Boolean that indicates whether to backup the dataset or not
        """
        self.input_data = input_data
        self.backup_data = None
        self.backup = backup

    @abstractmethod
    def transform(self) -> None:
        """
        Function that performs a tranformation to the dataset
        """
        pass

    def apply(self) -> None:
        """
        Backup the dataset if indicated and applies the transformation
        """
        if self.input_data is not None:
            # We first backup the data of the input object
            if self.backup:
                self.backup_data = copy.deepcopy(self.input_data)
            self.transform()
        else:
            raise ValueError("Input dataset object has not been provided")



import logging
from .field import Field
from .polarization import Polarization
from .spectral_window import SpectralWindow
from .visibility import Visibility


class MS:
    def __init__(self, _id: int = None, field: Field = None, polarization: Polarization = None,
                 spectral_window: SpectralWindow = None, visibilities: Visibility = None):
        """
        Class that represents the partitioned Measurement Set. Remember that dask-ms partitions the whole MS
        into Fields and Spectral Windows.

        Parameters
        ----------
        _id : int
             ID for each MS partition
        field : Field
               Field object of this MS partition
        polarization : Polarization
                      Polarization object for this MS partition
        spectral_window : SpectralWindow
                         Spectral Window object for this MS partition
        visibilities : Visibility
                      Visibility object for this MS partition
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self._id = _id
        self.field = field
        self.polarization = polarization
        self.spectral_window = spectral_window
        self.visibilities = visibilities



from astropy import units as u
from ..utils.analytical_functions import Gaussian2D
from .transformer import Transformer
from ..units.units_functions import check_units
from ..units.lambda_units import lambdas_equivalencies
from ..base.dataset import Dataset
from abc import ABC
from astropy.units import Quantity
import xarray as xr
import sys


class UVTaper(Transformer, Gaussian2D, ABC):
    def __init__(self, input_data: Dataset = None, amplitude: float = None, mu: Quantity = None,
                 sigma: Quantity = None, fwhm: Quantity = None, theta: Quantity = None):
        """
        Class that represents the UVTaper used in interferometry, the methods of this class will
        transform the weights of the input dataset according to the UVTaper features. UVTaper inherits from the
        Gaussian2D class.
        :param input_data: The input dataset to transform
        :param amplitude: Amplitude of the UVTaper
        :param mu: Center of the UVTaper
        :param sigma: Standard deviation of the UVTaper
        :param fwhm: FWHM of the UVTaper
        :param theta: Position Angle of the UVTaper
        """
        Transformer.__init__(self, input_data=input_data)

        # Check that sigmas and fwhm are in meters or units of lambdas
        if sigma is not None:
            if not check_units(sigma, u.m) and not check_units(sigma, u.lambdas) and not check_units(sigma, u.rad):
                raise ValueError("Sigma parameter for uvtaper must be whether in meters, units of lambdas or units of "
                                 "angle")
            else:
                # Make sure that units are in lambdas and not klambdas, Mlambdas, etc
                if check_units(sigma, u.lambdas):
                    sigma = sigma.to(u.lambdas)
                elif check_units(sigma, u.m):
                    sigma = sigma.to(u.m)
                else:
                    sigma = sigma.to(u.lambdas, equivalencies=lambdas_equivalencies())

        if fwhm is not None:
            if not check_units(fwhm, u.m) and not check_units(fwhm, u.lambdas) and not check_units(fwhm, u.rad):
                raise ValueError("FWHM parameter for uvtaper must be whether in meters, units of lambdas or units of "
                                 "angle")
            else:
                # Make sure that units are in lambdas and not klambdas, Mlambdas, etc
                if check_units(fwhm, u.lambdas):
                    fwhm = fwhm.to(u.lambdas)
                elif check_units(fwhm, u.m):
                    fwhm = fwhm.to(u.m)
                else:
                    fwhm = fwhm.to(u.lambdas, equivalencies=lambdas_equivalencies())

        # Check that theta is in units equivalent to radians

        if theta is not None:
            if not check_units(theta, u.rad):
                raise ValueError("Theta angle parameter for uvtaper must be in units of angles")

        Gaussian2D.__init__(self, amplitude, mu, sigma, theta, fwhm)

    def transform(self) -> None:
        """
        Evaluates the taper on the (u,v) positions and tapers the weights.
        """
        for ms in self.input_data.ms_list:
            if self.sigma.unit == u.m:
                u_xarray = ms.visibilities.uvw[:, 0]  # This is in meters
                v_xarray = ms.visibilities.uvw[:, 1]  # This is in meters
            else:
                uvw_lambdas = ms.visibilities.get_uvw_lambda(nu=ms.spectral_window.ref_frequency)
                u_xarray = uvw_lambdas[:, 0]  # This is in units of lambdas
                v_xarray = uvw_lambdas[:, 1]  # This is in units of lambdas

            u_xarray = u_xarray.astype('float32')  # Changing to float32 so the result from evaluate is a float32 array
            v_xarray = v_xarray.astype('float32')  # Changing to float32 so the result from evaluate is a float32 array
            gauss = xr.apply_ufunc(self.evaluate, u_xarray, v_xarray, dask="parallelized",
                                   output_dtypes=[ms.visibilities.weight.dtype])
            ms.visibilities.weight *= gauss



import xarray as xr
import dask.array as da
from astropy.units import Quantity
from astropy.units.equivalencies import Equivalency
from typing import Union


def xarray_unit_conversion(array: xr.DataArray = None, unit: Quantity = None, equivalencies: Equivalency = None,
                           dask: str = "parallelized") -> xr.DataArray:
    return xr.apply_ufunc(lambda x: x.to(unit, equivalencies=equivalencies), array, dask=dask,
                          output_dtypes=[array.dtype])


def dask_unit_conversion(array: da.core.Array = None, unit: Quantity = None, equivalencies: Equivalency = None) -> da.core.Array:
    return da.map_blocks(lambda x: x.to(unit, equivalencies), array, dtype=array.dtype)


def array_unit_conversion(array: Union[xr.DataArray, da.core.Array] = None, unit: Quantity = None, equivalencies: Equivalency = None) -> Union[xr.DataArray, da.core.Array]:
    if isinstance(array, xr.DataArray):
        return xarray_unit_conversion(array, unit, equivalencies)
    elif isinstance(array, da.core.Array):
        return dask_unit_conversion(array, unit, equivalencies)
    else:
        raise TypeError("The array is not a dask nor xarray type")


def check_units(var: Quantity, unit: Quantity) -> bool:
    """
    Checks if a variable unit is equivalent to some other unit
    :param var: The variable
    :param unit: The unit
    :return: True if the variable unit is equivalent, False if not
    """
    if var.unit.is_equivalent(unit):
        return True
    else:
        return False



import numpy as np
import astropy.units as u
from astropy.units import Quantity
import dask.array as da
import xarray as xr
from astropy.constants import c
from .antenna import Antenna
from .baseline import Baseline
from .ms import MS
from ..reconstruction import PSF
from ..units.lambda_units import lambdas_equivalencies
from ..units import array_unit_conversion
from typing import List, Union
from more_itertools import locate
import logging


def calc_beam_size(s_uu, s_vv, s_uv) -> tuple:
    """

    Parameters
    ----------
    s_uu : float
          Weighted sum of u^2.
    s_vv : float
          Weighted sum of v^2.
    s_uv : float
          Weighted sum of u*v.

    Returns
    -------
    tuple
        Beam major, minor and position angle in radians.
    """
    with np.errstate(divide='ignore'):
        uv_squared = s_uv ** 2
        uu_minus_vv = s_uu - s_vv
        uu_plus_vv = s_uu + s_vv
        sqrt_in = np.sqrt(4.0 * uv_squared + (uu_minus_vv ** 2))
        bmaj = 2.0 * np.sqrt(np.log(2.0)) / np.pi / np.sqrt(uu_plus_vv - sqrt_in)  # Major axis in radians
        bmin = 2.0 * np.sqrt(np.log(2.0)) / np.pi / np.sqrt(uu_plus_vv + sqrt_in)  # Minor axis in radians
        bpa = -0.5 * np.arctan2(2.0 * s_uv, uu_minus_vv)  # Angle in radians
        return bmaj * u.rad, bmin * u.rad, bpa * u.rad


class Dataset:
    def __init__(self, antenna: Antenna = None, baseline: Baseline = None,
                 spectral_window_dataset: List[xr.Dataset] = None, ms_list: List[MS] = None,
                 psf: Union[PSF, List[PSF]] = None, feed_kind: str = None):
        """

        Parameters
        ----------
        antenna : Full Antenna object
        baseline : Full Baseline object
        spectral_window_dataset : Full Spectral Window dataset
        ms_list : List of separated MS
        psf : PSF object of the dataset, or list of PSF objects for each one the Stokes parameters
        feed_kind : Kind of feed of the dataset. E.g "linear", "circular".
        """

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.antenna = antenna
        self.baseline = baseline
        self.spectral_window_dataset = spectral_window_dataset
        self.ms_list = ms_list
        self.psf = psf
        self.feed_kind = feed_kind

        self.max_nu = 0.0 * u.Hz  # Maximum frequency in Hz
        self.min_nu = 0.0 * u.Hz  # Minimum frequency in Hz
        self.ref_nu = 0.0 * u.Hz  # Reference frequency in Hz

        self.lambda_min = 0.0 * u.m  # Minimum wavelength in meters
        self.lambda_max = 0.0 * u.m  # Maximum wavelength in meters
        self.lambda_ref = 0.0 * u.m  # Reference wavelength in meters

        self.max_baseline = 0.0 * u.m  # Maximum baseline in meters
        self.min_baseline = 0.0 * u.m  # Minimum baseline in meters

        self.max_antenna_diameter = 0.0 * u.m  # Maximum antenna diameter in meters
        self.min_antenna_diameter = 0.0 * u.m  # Minimum antenna diameter in meters

        self.theo_resolution = 0.0 * u.rad  # Theoretical resolution in radians
        self.fov = 0.0 * u.rad  # Field-of-view in radians

        self.ndatasets = 0

        self.corr_weight_sum = None

        if spectral_window_dataset is not None:
            max_freqs = []
            min_freqs = []
            ref_freqs = []
            for spw in spectral_window_dataset:
                max_freqs.append(spw.CHAN_FREQ.max().data)
                min_freqs.append(spw.CHAN_FREQ.min().data)
                ref_freqs.append(spw.REF_FREQUENCY.data)

            max_freqs = list(da.compute(*max_freqs))
            min_freqs = list(da.compute(*min_freqs))

            self.max_nu = max(max_freqs) * u.Hz
            self.min_nu = min(min_freqs) * u.Hz
            self.ref_nu = np.median(np.array(ref_freqs)) * u.Hz

            self.lambda_min = c / self.max_nu
            self.lambda_min = self.lambda_min.to(u.m)
            self.lambda_max = c / self.min_nu
            self.lambda_max = self.lambda_max.to(u.m)
            self.lambda_ref = c / self.ref_nu
            self.lambda_ref = self.lambda_ref.to(u.m)

        if antenna is not None:
            self.max_antenna_diameter = antenna.max_diameter
            self.min_antenna_diameter = antenna.min_diameter

        if baseline is not None:
            self.max_baseline = baseline.max_baseline
            self.min_baseline = baseline.min_baseline

        if antenna is not None and baseline is not None and spectral_window_dataset is not None:
            self.theo_resolution = (self.lambda_min / self.max_baseline) * u.rad
            self.fov = (self.lambda_max / self.max_antenna_diameter) * u.rad

        if ms_list is not None:
            self.ndatasets = len(self.ms_list)
            self.check_feed()

            if self.psf is None:
                self.calculate_psf()

    def check_feed(self) -> None:
        """
        Function to check if a feed of the dataset is
        linear, circular or mixed
        """
        feed_list = []
        for ms in self.ms_list:
            feed_list.append(ms.polarization.feed_kind)
        if all(x == "linear" for x in feed_list):
            self.feed_kind = "linear"
        elif all(x == "circular" for x in feed_list):
            self.feed_kind = "circular"
        else:
            self.feed_kind = "mixed"

    def max_ncorrs(self) -> int:
        ncorrs = []
        for ms in self.ms_list:
            ncorrs.append(ms.polarization.ncorrs)
        return max(ncorrs)

    def calculate_weights_sum(self) -> None:

        if self.feed_kind == "linear":
            weight_corr_dict = {'XX': 0.0, 'YX': 0.0, 'XY': 0.0, 'YY': 0.0}
        elif self.feed_kind == "circular":
            weight_corr_dict = {'LL': 0.0, 'RL': 0.0, 'LR': 0.0, 'RR': 0.0}
        else:
            weight_corr_dict = {'XX': 0.0, 'YX': 0.0, 'XY': 0.0, 'YY': 0.0, 'LL': 0.0, 'RL': 0.0, 'LR': 0.0, 'RR': 0.0}

        for ms in self.ms_list:
            weight = ms.visibilities.weight
            flag = ms.visibilities.flag
            ncorrs = ms.polarization.ncorrs
            nchans = ms.spectral_window.nchans
            corr_names = ms.polarization.corrs_names
            weight_broadcast = da.repeat(weight.data, nchans, axis=0)
            flag = flag.data.reshape((len(flag.data) * nchans, ncorrs))
            weight_broadcast[flag] = 0.0

            weight_sum = da.sum(weight_broadcast, axis=0)

            for idx_corr in range(0, ncorrs):
                weight_corr_dict[corr_names[idx_corr]] += weight_sum[idx_corr]

        weight_corr_dict = da.compute(weight_corr_dict)[0]

        self.corr_weight_sum = weight_corr_dict

    def calculate_psf(self, stokes: Union[List[str], str] = None) -> None:
        """
        Function that calculates the PSF properties (bmaj, bmin and bpa) analytically for different stokes using
        (u,v) positions and the weights
        """
        if stokes is None:
            stokes = ["I", "Q", "U", "V"]
        else:
            if isinstance(stokes, str):
                stokes = stokes.split(',')
        self.calculate_weights_sum()

        idx_I = list(locate(stokes, lambda x: x == "I"))
        idx_Q = list(locate(stokes, lambda x: x == "Q"))
        idx_U = list(locate(stokes, lambda x: x == "U"))
        idx_V = list(locate(stokes, lambda x: x == "V"))

        nstokes = len(stokes)

        s_uu = da.zeros(nstokes, dtype=np.float64)
        s_vv = da.zeros(nstokes, dtype=np.float64)
        s_uv = da.zeros(nstokes, dtype=np.float64)
        weight_sum = np.zeros(nstokes, dtype=np.float32)

        for ms in self.ms_list:
            chans = ms.spectral_window.chans.compute()
            nchans = ms.spectral_window.nchans
            uvw = ms.visibilities.uvw
            weight = ms.visibilities.weight
            flag = ms.visibilities.flag
            ncorrs = ms.polarization.ncorrs
            corr_names = ms.polarization.corrs_names

            uvw_broadcast = da.tile(uvw, nchans).reshape((len(uvw), nchans, 3))
            chans_broadcast = chans[np.newaxis, :, np.newaxis]

            uvw_lambdas = array_unit_conversion(array=uvw_broadcast, unit=u.lambdas,
                                                equivalencies=lambdas_equivalencies(restfreq=chans_broadcast))

            uvw_lambdas = da.map_blocks(lambda x: x.value, uvw_lambdas, dtype=np.float64)

            uvw_lambdas = uvw_lambdas.reshape((len(uvw_lambdas) * nchans, 3))
            weight_broadcast = da.repeat(weight.data, nchans, axis=0)
            flag = flag.data.reshape((len(flag.data) * nchans, ncorrs))
            weight_broadcast[flag] = 0.0

            idx_corrs = [key for key, value in corr_names.items()]
            _u = uvw_lambdas[:, 0, np.newaxis]
            _v = uvw_lambdas[:, 1, np.newaxis]

            _s_uu = da.sum(weight_broadcast * _u ** 2, axis=0)
            _s_vv = da.sum(weight_broadcast * _v ** 2, axis=0)
            _s_uv = da.sum(weight_broadcast * _u * _v, axis=0)

            # _s_uu = _s_uu.compute().value
            # _s_vv = _s_vv.compute().value
            # _s_uv = _s_uv.compute().value

            for i in idx_corrs:
                if corr_names[i] == "XX":
                    if idx_I:
                        s_uu[idx_I[0]] += _s_uu[i]
                        s_vv[idx_I[0]] += _s_vv[i]
                        s_uv[idx_I[0]] += _s_uv[i]
                        weight_sum[idx_I[0]] += self.corr_weight_sum[corr_names[i]]

                    if idx_Q:
                        s_uu[idx_Q[0]] += _s_uu[i]
                        s_vv[idx_Q[0]] += _s_vv[i]
                        s_uv[idx_Q[0]] += _s_uv[i]
                        weight_sum[idx_Q[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "XY":
                    if idx_U:
                        s_uu[idx_U[0]] += _s_uu[i]
                        s_vv[idx_U[0]] += _s_vv[i]
                        s_uv[idx_U[0]] += _s_uv[i]
                        weight_sum[idx_U[0]] += self.corr_weight_sum[corr_names[i]]

                    if idx_V:
                        s_uu[idx_V[0]] += _s_uu[i]
                        s_vv[idx_V[0]] += _s_vv[i]
                        s_uv[idx_V[0]] += _s_uv[i]
                        weight_sum[idx_V[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "YX":
                    if idx_U:
                        s_uu[idx_U[0]] += _s_uu[i]
                        s_vv[idx_U[0]] += _s_vv[i]
                        s_uv[idx_U[0]] += _s_uv[i]
                        weight_sum[idx_U[0]] += self.corr_weight_sum[corr_names[i]]

                    if idx_V:
                        s_uu[idx_V[0]] += _s_uu[i]
                        s_vv[idx_V[0]] += _s_vv[i]
                        s_uv[idx_V[0]] += _s_uv[i]
                        weight_sum[idx_V[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "YY":
                    if idx_I:
                        s_uu[idx_I[0]] += _s_uu[i]
                        s_vv[idx_I[0]] += _s_vv[i]
                        s_uv[idx_I[0]] += _s_uv[i]
                        weight_sum[idx_I[0]] += self.corr_weight_sum[corr_names[i]]
                    if idx_Q:
                        s_uu[idx_Q[0]] += _s_uu[i]
                        s_vv[idx_Q[0]] += _s_vv[i]
                        s_uv[idx_Q[0]] += _s_uv[i]
                        weight_sum[idx_Q[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "LL":
                    if idx_I:
                        s_uu[idx_I[0]] += _s_uu[i]
                        s_vv[idx_I[0]] += _s_vv[i]
                        s_uv[idx_I[0]] += _s_uv[i]
                        weight_sum[idx_I[0]] += self.corr_weight_sum[corr_names[i]]
                    if idx_V:
                        s_uu[idx_V[0]] += _s_uu[i]
                        s_vv[idx_V[0]] += _s_vv[i]
                        s_uv[idx_V[0]] += _s_uv[i]
                        weight_sum[idx_V[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "LR":
                    if idx_Q:
                        s_uu[idx_Q[0]] += _s_uu[i]
                        s_vv[idx_Q[0]] += _s_vv[i]
                        s_uv[idx_Q[0]] += _s_uv[i]
                        weight_sum[idx_Q[0]] += self.corr_weight_sum[corr_names[i]]
                    if idx_U:
                        s_uu[idx_U[0]] += _s_uu[i]
                        s_vv[idx_U[0]] += _s_vv[i]
                        s_uv[idx_U[0]] += _s_uv[i]
                        weight_sum[idx_U[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "RL":
                    if idx_Q:
                        s_uu[idx_Q[0]] += _s_uu[i]
                        s_vv[idx_Q[0]] += _s_vv[i]
                        s_uv[idx_Q[0]] += _s_uv[i]
                        weight_sum[idx_Q[0]] += self.corr_weight_sum[corr_names[i]]

                    if idx_U:
                        s_uu[idx_U[0]] += _s_uu[i]
                        s_vv[idx_U[0]] += _s_vv[i]
                        s_uv[idx_U[0]] += _s_uv[i]
                        weight_sum[idx_U[0]] += self.corr_weight_sum[corr_names[i]]
                elif corr_names[i] == "RR":
                    if idx_I:
                        s_uu[idx_I[0]] += _s_uu[i]
                        s_vv[idx_I[0]] += _s_vv[i]
                        s_uv[idx_I[0]] += _s_uv[i]
                        weight_sum[idx_I[0]] += self.corr_weight_sum[corr_names[i]]

                    if idx_V:
                        s_uu[idx_V[0]] += _s_uu[i]
                        s_vv[idx_V[0]] += _s_vv[i]
                        s_uv[idx_V[0]] += _s_uv[i]
                        weight_sum[idx_V[0]] += self.corr_weight_sum[corr_names[i]]
                else:
                    raise ValueError("The correlation does not exist")

        s_uu[weight_sum > 0.0] /= weight_sum[weight_sum > 0.0]
        s_vv[weight_sum > 0.0] /= weight_sum[weight_sum > 0.0]
        s_uv[weight_sum > 0.0] /= weight_sum[weight_sum > 0.0]

        s_uu, s_vv, s_uv = da.compute(s_uu, s_vv, s_uv)
        bmaj, bmin, bpa = calc_beam_size(s_uu, s_vv, s_uv)

        psf_objects = []
        for i in range(0, nstokes):
            if weight_sum[i] > 0.0:
                sigma = Quantity([bmaj[i], bmin[i]])
                psf_objects.append(PSF(sigma=sigma, theta=bpa[i]))
            else:
                psf_objects.append(None)

        if len(psf_objects) == 1:
            self.psf = psf_objects[0]
        else:
            self.psf = psf_objects
